@@ -28,6 +28,7 @@
 #include <unistd.h>
 #include <webkit2/webkit2.h>
 
+#include "../version.h"
 #include "ascii.h"
 #include "command.h"
 #include "completion.h"
@@ -45,6 +46,7 @@
 #include "shortcut.h"
 #include "util.h"
 #include "autocmd.h"
+#include "file-storage.h"
 
 static void client_destroy(Client *c);
 static Client *client_new(WebKitWebView *webview);
@@ -62,6 +64,9 @@ static void on_webctx_download_started(WebKitWebContext *webctx,
 static void on_webctx_init_web_extension(WebKitWebContext *webctx, gpointer data);
 static gboolean on_webdownload_decide_destination(WebKitDownload *download,
         gchar *suggested_filename, Client *c);
+static void on_webdownload_response_received(WebKitDownload *download,
+        GParamSpec *ps, Client *c);
+static void spawn_download_command(Client *c, WebKitURIResponse *response);
 static void on_webdownload_failed(WebKitDownload *download,
         GError *error, Client *c);
 static void on_webdownload_finished(WebKitDownload *download, Client *c);
@@ -72,6 +77,9 @@ static WebKitWebView *on_webview_create(WebKitWebView *webview,
         WebKitNavigationAction *navact, Client *c);
 static gboolean on_webview_decide_policy(WebKitWebView *webview,
         WebKitPolicyDecision *dec, WebKitPolicyDecisionType type, Client *c);
+static void decide_navigation_action(Client *c, WebKitPolicyDecision *dec);
+static void decide_new_window_action(Client *c, WebKitPolicyDecision *dec);
+static void decide_response(Client *c, WebKitPolicyDecision *dec);
 static void on_webview_load_changed(WebKitWebView *webview,
         WebKitLoadEvent event, Client *c);
 static void on_webview_mouse_target_changed(WebKitWebView *webview,
@@ -86,6 +94,8 @@ static void on_webview_ready_to_show(WebKitWebView *webview, Client *c);
 static gboolean on_webview_web_process_crashed(WebKitWebView *webview, Client *c);
 static gboolean on_webview_authenticate(WebKitWebView *webview,
         WebKitAuthenticationRequest *request, Client *c);
+static gboolean on_webview_enter_fullscreen(WebKitWebView *webview, Client *c);
+static gboolean on_webview_leave_fullscreen(WebKitWebView *webview, Client *c);
 static gboolean on_window_delete_event(GtkWidget *window, GdkEvent *event, Client *c);
 static void on_window_destroy(GtkWidget *window, Client *c);
 static gboolean quit(Client *c);
@@ -108,6 +118,8 @@ static void on_script_message_focus(WebKitUserContentManager *manager,
         WebKitJavascriptResult *res, gpointer data);
 static gboolean profileOptionArgFunc(const gchar *option_name,
         const gchar *value, gpointer data, GError **error);
+static gboolean autocmdOptionArgFunc(const gchar *option_name,
+        const gchar *value, gpointer data, GError **error);
 
 struct Vimb vb;
 
@@ -632,6 +644,39 @@ void vb_statusbar_update(Client *c)
     g_string_free(status, TRUE);
 }
 
+/**
+ * Show the given url on the left of statusbar.
+ */
+void vb_statusbar_show_hover_url(Client *c, VbLinkType type, const char *uri)
+{
+    char *sanitized_uri,
+         *msg;
+    const char *type_label;
+
+    /* No uri given - show the current URI. */
+    if (!uri || !*uri) {
+        update_urlbar(c);
+        return;
+    }
+
+    switch (type) {
+        case LINK_TYPE_LINK:
+            type_label = "Link: ";
+            break;
+        case LINK_TYPE_IMAGE:
+            type_label = "Image: ";
+            break;
+        default:
+            return;
+    }
+
+    sanitized_uri = util_sanitize_uri(uri);
+    msg           = g_strconcat(type_label, uri, NULL);
+    gtk_label_set_text(GTK_LABEL(c->statusbar.left), msg);
+    g_free(msg);
+    g_free(sanitized_uri);
+}
+
 /**
  * Destroys given client and removed it from client queue. If no client is
  * there in queue, quit the gtk main loop.
@@ -801,7 +846,9 @@ static GtkWidget *create_window(Client *c)
         window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
         gtk_window_set_role(GTK_WINDOW(window), PROJECT_UCFIRST);
         gtk_window_set_default_size(GTK_WINDOW(window), WIN_WIDTH, WIN_HEIGHT);
-        gtk_window_maximize(GTK_WINDOW(window));
+        if (!vb.no_maximize) {
+            gtk_window_maximize(GTK_WINDOW(window));
+        }
     }
 
     g_object_connect(
@@ -982,7 +1029,10 @@ static void spawn_new_instance(const char *uri)
 #ifndef FEATURE_NO_XEMBED
         + (vb.embed ? 2 : 0)
 #endif
-        + (vb.profile ? 2 : 0),
+        + (vb.incognito ? 1 : 0)
+        + (vb.profile ? 2 : 0)
+        + (vb.no_maximize ? 1 : 0)
+        + g_slist_length(vb.cmdargs) * 2,
         sizeof(char *)
     );
 
@@ -1000,10 +1050,20 @@ static void spawn_new_instance(const char *uri)
         cmd[i++] = xid;
     }
 #endif
+    if (vb.incognito) {
+        cmd[i++] = "-i";
+    }
     if (vb.profile) {
         cmd[i++] = "-p";
         cmd[i++] = vb.profile;
     }
+    if (vb.no_maximize) {
+        cmd[i++] = "--no-maximize";
+    }
+    for (GSList *l = vb.cmdargs; l; l = l->next) {
+        cmd[i++] = "-C";
+        cmd[i++] = l->data;
+    }
     cmd[i++] = (char*)uri;
     cmd[i++] = NULL;
 
@@ -1025,15 +1085,19 @@ static void on_webctx_download_started(WebKitWebContext *webctx,
     autocmd_run(c, AU_DOWNLOAD_STARTED, uri, NULL);
 #endif
 
-    g_signal_connect(download, "decide-destination", G_CALLBACK(on_webdownload_decide_destination), c);
-    g_signal_connect(download, "failed", G_CALLBACK(on_webdownload_failed), c);
-    g_signal_connect(download, "finished", G_CALLBACK(on_webdownload_finished), c);
-    g_signal_connect(download, "received-data", G_CALLBACK(on_webdownload_received_data), c);
+    if (GET_BOOL(c, "download-use-external")) {
+        g_signal_connect(download, "notify::response", G_CALLBACK(on_webdownload_response_received), c);
+    } else {
+        g_signal_connect(download, "decide-destination", G_CALLBACK(on_webdownload_decide_destination), c);
+        g_signal_connect(download, "failed", G_CALLBACK(on_webdownload_failed), c);
+        g_signal_connect(download, "finished", G_CALLBACK(on_webdownload_finished), c);
+        g_signal_connect(download, "received-data", G_CALLBACK(on_webdownload_received_data), c);
 
-    c->state.downloads = g_list_append(c->state.downloads, download);
+        c->state.downloads = g_list_append(c->state.downloads, download);
 
-    /* to reflect the correct download count */
-    vb_statusbar_update(c);
+        /* to reflect the correct download count */
+        vb_statusbar_update(c);
+    }
 }
 
 /**
@@ -1044,20 +1108,20 @@ static void on_webctx_init_web_extension(WebKitWebContext *webctx, gpointer data
     const char *name;
     GVariant *vdata;
 
-#ifdef DEBUG
+#if (CHECK_WEBEXTENSION_ON_STARTUP)
     char *extension = g_build_filename(EXTENSIONDIR,  "webext_main.so", NULL);
     if (!g_file_test(extension, G_FILE_TEST_IS_REGULAR)) {
-        g_warning("Cannot access web extension %s", extension);
+        g_error("Cannot access web extension %s", extension);
     }
     g_free(extension);
 #endif
 
-    /* Setup the extension directory. */
-    webkit_web_context_set_web_extensions_directory(webctx, EXTENSIONDIR);
-
     name  = ext_proxy_init();
     vdata = g_variant_new("(ms)", name);
     webkit_web_context_set_web_extensions_initialization_user_data(webctx, vdata);
+
+    /* Setup the extension directory. */
+    webkit_web_context_set_web_extensions_directory(webctx, EXTENSIONDIR);
 }
 
 /**
@@ -1075,6 +1139,48 @@ static gboolean on_webdownload_decide_destination(WebKitDownload *download,
     return vb_download_set_destination(c, download, suggested_filename, NULL);
 }
 
+static void on_webdownload_response_received(WebKitDownload *download,
+        GParamSpec *ps, Client *c)
+{
+    spawn_download_command(c, webkit_download_get_response(download));
+    webkit_download_cancel(download);
+}
+
+static void spawn_download_command(Client *c, WebKitURIResponse *response)
+{
+    char *cmd;
+    char **argv, **envp;
+    int argc;
+    GError *error = NULL;
+
+    cmd = g_strdup_printf(GET_CHAR(c, "download-command"),
+            webkit_uri_response_get_uri(response));
+
+    if (!g_shell_parse_argv(cmd, &argc, &argv, &error)) {
+        g_warning("Could not parse download-command '%s': %s",
+                cmd,
+                error->message);
+        g_error_free(error);
+        g_free(cmd);
+        return;
+    }
+
+    envp = g_get_environ();
+    envp = g_environ_setenv(envp, "VIMB_DOWNLOAD_PATH",
+            GET_CHAR(c, "download-path"), TRUE);
+
+    if (g_spawn_async(NULL, argv, envp, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, &error)) {
+        vb_echo(c, MSG_NORMAL, FALSE, "Download started");
+    } else {
+        vb_echo(c, MSG_ERROR, TRUE, "Could not start download");
+        g_warning("%s", error->message);
+        g_clear_error(&error);
+    }
+    g_free(cmd);
+    g_strfreev(envp);
+    g_strfreev(argv);
+}
+
 /**
  * Callback for the webkit download failed signal.
  * This signal is emitted when an error occurs during the download operation.
@@ -1192,7 +1298,7 @@ static void on_webdownload_received_data(WebKitDownload *download,
  */
 static void on_webview_close(WebKitWebView *webview, Client *c)
 {
-    client_destroy(c);
+    gtk_widget_destroy(c->window);
 }
 
 /**
@@ -1202,6 +1308,14 @@ static void on_webview_close(WebKitWebView *webview, Client *c)
 static WebKitWebView *on_webview_create(WebKitWebView *webview,
         WebKitNavigationAction *navact, Client *c)
 {
+    WebKitURIRequest *req;
+    if (c->config.prevent_newwindow) {
+        req = webkit_navigation_action_get_request(navact);
+        vb_load_uri(c, &(Arg){TARGET_CURRENT, (char*)webkit_uri_request_get_uri(req)});
+
+        return NULL;
+    }
+
     Client *new = client_new(webview);
 
     return new->webview;
@@ -1215,75 +1329,115 @@ static WebKitWebView *on_webview_create(WebKitWebView *webview,
 static gboolean on_webview_decide_policy(WebKitWebView *webview,
         WebKitPolicyDecision *dec, WebKitPolicyDecisionType type, Client *c)
 {
-    guint status, button, mod;
-    WebKitNavigationAction *a;
-    WebKitURIRequest *req;
-    WebKitURIResponse *res;
-    const char *uri;
-
     switch (type) {
         case WEBKIT_POLICY_DECISION_TYPE_NAVIGATION_ACTION:
-            a      = webkit_navigation_policy_decision_get_navigation_action(WEBKIT_NAVIGATION_POLICY_DECISION(dec));
-            req    = webkit_navigation_action_get_request(a);
-            button = webkit_navigation_action_get_mouse_button(a);
-            mod    = webkit_navigation_action_get_modifiers(a);
-            uri    = webkit_uri_request_get_uri(req);
-
-            /* Try to handle with specific protocol handler. */
-            if (handler_handle_uri(c->handler, uri)) {
-                webkit_policy_decision_ignore(dec);
-                return TRUE;
-            }
-            /* Spawn new instance if the new win flag is set on the mode, or
-             * the navigation was triggered by CTRL-LeftMouse or MiddleMouse. */
-            if ((c->mode->flags & FLAG_NEW_WIN)
-                || (webkit_navigation_action_get_navigation_type(a) == WEBKIT_NAVIGATION_TYPE_LINK_CLICKED
-                    && (button == 2 || (button == 1 && mod & GDK_CONTROL_MASK)))) {
-
-                /* Remove the FLAG_NEW_WIN after the first use. */
-                c->mode->flags &= ~FLAG_NEW_WIN;
-
-                webkit_policy_decision_ignore(dec);
-                spawn_new_instance(uri);
-                return TRUE;
-            }
-            return FALSE;
+            decide_navigation_action(c, dec);
+            break;
 
         case WEBKIT_POLICY_DECISION_TYPE_NEW_WINDOW_ACTION:
-            a   = webkit_navigation_policy_decision_get_navigation_action(WEBKIT_NAVIGATION_POLICY_DECISION(dec));
-
-            /* Ignore opening new window if this was started without user gesture. */
-            if (!webkit_navigation_action_is_user_gesture(a)) {
-                webkit_policy_decision_ignore(dec);
-                return TRUE;
-            }
-
-            if (webkit_navigation_action_get_navigation_type(a) == WEBKIT_NAVIGATION_TYPE_LINK_CLICKED) {
-                webkit_policy_decision_ignore(dec);
-                /* This is triggered on link click for links with
-                 * target="_blank". Maybe it should be configurable if the
-                 * page is opened as tab or a new instance. */
-                req = webkit_navigation_action_get_request(a);
-                spawn_new_instance(webkit_uri_request_get_uri(req));
-                return TRUE;
-            }
-            return FALSE;
+            decide_new_window_action(c, dec);
+            break;
 
         case WEBKIT_POLICY_DECISION_TYPE_RESPONSE:
-            res    = webkit_response_policy_decision_get_response(WEBKIT_RESPONSE_POLICY_DECISION(dec));
-            status = webkit_uri_response_get_status_code(res);
-
-            if (!webkit_response_policy_decision_is_mime_type_supported(WEBKIT_RESPONSE_POLICY_DECISION(dec))
-                    && (SOUP_STATUS_IS_SUCCESSFUL(status) || status == SOUP_STATUS_NONE)) {
-
-                webkit_policy_decision_download(dec);
-
-                return TRUE;
-            }
-            return FALSE;
+            decide_response(c, dec);
+            break;
 
         default:
-            return FALSE;
+            webkit_policy_decision_ignore(dec);
+            break;
+    }
+
+    return TRUE;
+}
+
+static void decide_navigation_action(Client *c, WebKitPolicyDecision *dec)
+{
+    guint button, mod;
+    WebKitNavigationAction *a;
+    WebKitURIRequest *req;
+    const char *uri;
+
+    a   = webkit_navigation_policy_decision_get_navigation_action(WEBKIT_NAVIGATION_POLICY_DECISION(dec));
+    req = webkit_navigation_action_get_request(a);
+    uri = webkit_uri_request_get_uri(req);
+
+    /* Try to handle with specific protocol handler. */
+    if (handler_handle_uri(c->handler, uri)) {
+        webkit_policy_decision_ignore(dec);
+        return;
+    }
+
+    button = webkit_navigation_action_get_mouse_button(a);
+    mod    = webkit_navigation_action_get_modifiers(a);
+    /* Spawn new instance if the new win flag is set on the mode, or the
+     * navigation was triggered by CTRL-LeftMouse or MiddleMouse. */
+    if ((c->mode->flags & FLAG_NEW_WIN)
+        || (webkit_navigation_action_get_navigation_type(a) == WEBKIT_NAVIGATION_TYPE_LINK_CLICKED
+            && (button == 2 || (button == 1 && mod & GDK_CONTROL_MASK)))) {
+
+        /* Remove the FLAG_NEW_WIN after the first use. */
+        c->mode->flags &= ~FLAG_NEW_WIN;
+
+        webkit_policy_decision_ignore(dec);
+        spawn_new_instance(uri);
+    } else {
+#ifdef FEATURE_AUTOCMD
+        if (strcmp(uri, "about:blank"))
+            autocmd_run(c, AU_LOAD_STARTING, uri, NULL);
+#endif
+        webkit_policy_decision_use(dec);
+    }
+}
+
+static void decide_new_window_action(Client *c, WebKitPolicyDecision *dec)
+{
+    WebKitNavigationAction *a;
+    WebKitURIRequest *req;
+
+    a = webkit_navigation_policy_decision_get_navigation_action(WEBKIT_NAVIGATION_POLICY_DECISION(dec));
+
+    switch (webkit_navigation_action_get_navigation_type(a)) {
+        case WEBKIT_NAVIGATION_TYPE_LINK_CLICKED:   /* fallthrough */
+        case WEBKIT_NAVIGATION_TYPE_FORM_SUBMITTED: /* fallthrough */
+        case WEBKIT_NAVIGATION_TYPE_BACK_FORWARD:   /* fallthrough */
+        case WEBKIT_NAVIGATION_TYPE_RELOAD:         /* fallthrough */
+        case WEBKIT_NAVIGATION_TYPE_FORM_RESUBMITTED:
+            /* This is triggered on link click for links with target="_blank".
+             * Maybe it should be configurable if the page is opened as tab or
+             * a new instance. Ignore opening new window if this was started
+             * without user gesture. */
+            if (webkit_navigation_action_is_user_gesture(a)) {
+                req = webkit_navigation_action_get_request(a);
+                if (c->config.prevent_newwindow) {
+                    /* Load the uri into the browser instance. */
+                    vb_load_uri(c, &(Arg){TARGET_CURRENT, (char*)webkit_uri_request_get_uri(req)});
+                } else {
+                    spawn_new_instance(webkit_uri_request_get_uri(req));
+                }
+            }
+            break;
+
+        case WEBKIT_NAVIGATION_TYPE_OTHER: /* fallthrough */
+        default:
+            break;
+    }
+    webkit_policy_decision_ignore(dec);
+}
+
+static void decide_response(Client *c, WebKitPolicyDecision *dec)
+{
+    guint status;
+    WebKitURIResponse *res;
+
+    res    = webkit_response_policy_decision_get_response(WEBKIT_RESPONSE_POLICY_DECISION(dec));
+    status = webkit_uri_response_get_status_code(res);
+
+    if (webkit_response_policy_decision_is_mime_type_supported(WEBKIT_RESPONSE_POLICY_DECISION(dec))) {
+        webkit_policy_decision_use(dec);
+    } else if (SOUP_STATUS_IS_SUCCESSFUL(status) || status == SOUP_STATUS_NONE) {
+        webkit_policy_decision_download(dec);
+    } else {
+        webkit_policy_decision_ignore(dec);
     }
 }
 
@@ -1291,18 +1445,25 @@ static void on_webview_load_changed(WebKitWebView *webview,
         WebKitLoadEvent event, Client *c)
 {
     GTlsCertificateFlags tlsflags;
+    const char *raw_uri;
     char *uri = NULL;
 
+    raw_uri = webkit_web_view_get_uri(webview);
+    if (raw_uri) {
+        uri = util_sanitize_uri(raw_uri);
+    }
+
     switch (event) {
         case WEBKIT_LOAD_STARTED:
-            uri = util_sanitize_uri(webkit_web_view_get_uri(webview));
 #ifdef FEATURE_AUTOCMD
-            autocmd_run(c, AU_LOAD_STARTED, uri, NULL);
+            autocmd_run(c, AU_LOAD_STARTED, raw_uri, NULL);
 #endif
             /* update load progress in statusbar */
             c->state.progress = 0;
             vb_statusbar_update(c);
-            set_title(c, uri);
+            if (uri) {
+                set_title(c, uri);
+            }
             /* Make sure hinting is cleared before the new page is loaded.
              * Without that vimb would still be in hinting mode after hinting
              * was started and some links was clicked my mouse. Even if there
@@ -1322,14 +1483,13 @@ static void on_webview_load_changed(WebKitWebView *webview,
              * or aborted the load will be commited. So this seems to be the
              * right place to remove the flag. */
             c->mode->flags &= ~FLAG_IGNORE_FOCUS;
-            uri = util_sanitize_uri(webkit_web_view_get_uri(webview));
 #ifdef FEATURE_AUTOCMD
-            autocmd_run(c, AU_LOAD_COMMITTED, uri, NULL);
+            autocmd_run(c, AU_LOAD_COMMITTED, raw_uri, NULL);
 #endif
             /* save the current URI in register % */
             vb_register_add(c, '%', uri);
             /* check if tls is on and the page is trusted */
-            if (g_str_has_prefix(uri, "https://")) {
+            if (uri && g_str_has_prefix(uri, "https://")) {
                 if (webkit_web_view_get_tls_info(webview, NULL, &tlsflags) && tlsflags) {
                     set_statusbar_style(c, STATUS_SSL_INVALID);
                 } else {
@@ -1349,12 +1509,11 @@ static void on_webview_load_changed(WebKitWebView *webview,
             break;
 
         case WEBKIT_LOAD_FINISHED:
-            uri = util_sanitize_uri(webkit_web_view_get_uri(webview));
 #ifdef FEATURE_AUTOCMD
-            autocmd_run(c, AU_LOAD_FINISHED, uri, NULL);
+            autocmd_run(c, AU_LOAD_FINISHED, raw_uri, NULL);
 #endif
             c->state.progress = 100;
-            if (strncmp(uri, "about:", 6)) {
+            if (uri && strncmp(uri, "about:", 6)) {
                 history_add(c, HISTORY_URL, uri, webkit_web_view_get_title(webview));
             }
             break;
@@ -1373,9 +1532,6 @@ static void on_webview_load_changed(WebKitWebView *webview,
 static void on_webview_mouse_target_changed(WebKitWebView *webview,
         WebKitHitTestResult *result, guint modifiers, Client *c)
 {
-    char *msg;
-    char *uri;
-
     /* Save the hitTestResult to have this later available for events that
      * don't support this. */
     if (c->state.hit_test_result) {
@@ -1384,20 +1540,14 @@ static void on_webview_mouse_target_changed(WebKitWebView *webview,
     c->state.hit_test_result = g_object_ref(result);
 
     if (webkit_hit_test_result_context_is_link(result)) {
-        uri = util_sanitize_uri(webkit_hit_test_result_get_link_uri(result));
-        msg = g_strconcat("Link: ", uri, NULL);
-        gtk_label_set_text(GTK_LABEL(c->statusbar.left), msg);
-        g_free(msg);
-        g_free(uri);
+        vb_statusbar_show_hover_url(c, LINK_TYPE_LINK,
+                webkit_hit_test_result_get_link_uri(result));
     } else if (webkit_hit_test_result_context_is_image(result)) {
-        uri = util_sanitize_uri(webkit_hit_test_result_get_image_uri(result));
-        msg = g_strconcat("Image: ", uri, NULL);
-        gtk_label_set_text(GTK_LABEL(c->statusbar.left), msg);
-        g_free(msg);
-        g_free(uri);
+        vb_statusbar_show_hover_url(c, LINK_TYPE_LINK,
+                webkit_hit_test_result_get_image_uri(result));
     } else {
         /* No link under cursor - show the current URI. */
-        update_urlbar(c);
+        vb_statusbar_show_hover_url(c, LINK_TYPE_NONE, NULL);
     }
 }
 
@@ -1436,6 +1586,7 @@ static void on_webview_notify_uri(WebKitWebView *webview, GParamSpec *pspec, Cli
     if (c->state.uri) {
         g_free(c->state.uri);
     }
+
     c->state.uri = util_sanitize_uri(webkit_web_view_get_uri(c->webview));
 
     update_urlbar(c);
@@ -1477,6 +1628,28 @@ static gboolean on_webview_authenticate(WebKitWebView *webview,
     return FALSE;
 }
 
+/**
+ * Callback in case JS calls element.webkitRequestFullScreen.
+ */
+static gboolean on_webview_enter_fullscreen(WebKitWebView *webview, Client *c)
+{
+    c->state.is_fullscreen = TRUE;
+    gtk_widget_hide(GTK_WIDGET(c->statusbar.box));
+    gtk_widget_set_visible(GTK_WIDGET(c->input), FALSE);
+    return FALSE;
+}
+
+/**
+ * Callback to restore the window state after entering fullscreen.
+ */
+static gboolean on_webview_leave_fullscreen(WebKitWebView *webview, Client *c)
+{
+    c->state.is_fullscreen = FALSE;
+    gtk_widget_show(GTK_WIDGET(c->statusbar.box));
+    gtk_widget_set_visible(GTK_WIDGET(c->input), TRUE);
+    return FALSE;
+}
+
 /**
  * Callback for window ::delete-event signal which is emitted if a user
  * requests that a toplevel window is closed. The default handler for this
@@ -1611,12 +1784,17 @@ static void vimb_cleanup(void)
     /* free memory of other components */
     util_cleanup();
 
+    for (i = 0; i < STORAGE_LAST; i++) {
+        file_storage_free(vb.storage[i]);
+    }
     for (i = 0; i < FILES_LAST; i++) {
         if (vb.files[i]) {
             g_free(vb.files[i]);
         }
     }
     g_free(vb.profile);
+
+    g_slist_free_full(vb.cmdargs, g_free);
 }
 #endif
 
@@ -1637,25 +1815,32 @@ static void vimb_setup(void)
         vb.files[FILES_CONFIG] = g_strdup(rp);
         free(rp);
     } else {
-        vb.files[FILES_CONFIG] = util_get_filepath(path, "config", FALSE);
+        vb.files[FILES_CONFIG] = g_build_filename(path, "config", NULL);
     }
 
     /* Setup those files that are use multiple time during runtime */
-    vb.files[FILES_CLOSED]     = util_get_filepath(path, "closed", TRUE);
-    vb.files[FILES_COOKIE]     = util_get_filepath(path, "cookies.db", TRUE);
-    vb.files[FILES_USER_STYLE] = util_get_filepath(path, "style.css", FALSE);
-    vb.files[FILES_SCRIPT]     = util_get_filepath(path, "scripts.js", FALSE);
-    vb.files[FILES_HISTORY]    = util_get_filepath(path, "history", TRUE);
-    vb.files[FILES_COMMAND]    = util_get_filepath(path, "command", TRUE);
-    vb.files[FILES_BOOKMARK]   = util_get_filepath(path, "bookmark", TRUE);
-    vb.files[FILES_QUEUE]      = util_get_filepath(path, "queue", TRUE);
-    vb.files[FILES_SEARCH]     = util_get_filepath(path, "search", TRUE);
+    if (!vb.incognito) {
+        vb.files[FILES_CLOSED] = g_build_filename(path, "closed", NULL);
+        vb.files[FILES_COOKIE] = g_build_filename(path, "cookies.db", NULL);
+    }
+    vb.files[FILES_BOOKMARK]   = g_build_filename(path, "bookmark", NULL);
+    vb.files[FILES_QUEUE]      = g_build_filename(path, "queue", NULL);
+    vb.files[FILES_SCRIPT]     = g_build_filename(path, "scripts.js", NULL);
+    vb.files[FILES_USER_STYLE] = g_build_filename(path, "style.css", NULL);
+
+    vb.storage[STORAGE_HISTORY]  = file_storage_new(path, "history", vb.incognito);
+    vb.storage[STORAGE_COMMAND]  = file_storage_new(path, "command", vb.incognito);
+    vb.storage[STORAGE_SEARCH]   = file_storage_new(path, "search", vb.incognito);
     g_free(path);
 
     /* Use seperate rendering processed for the webview of the clients in the
      * current instance. This must be called as soon as possible according to
      * the documentation. */
-    ctx = webkit_web_context_get_default();
+    if (vb.incognito) {
+        ctx = webkit_web_context_new_ephemeral();
+    } else {
+        ctx = webkit_web_context_get_default();
+    }
     webkit_web_context_set_process_model(ctx, WEBKIT_PROCESS_MODEL_MULTIPLE_SECONDARY_PROCESSES);
     webkit_web_context_set_cache_model(ctx, WEBKIT_CACHE_MODEL_WEB_BROWSER);
 
@@ -1677,7 +1862,7 @@ static void vimb_setup(void)
     vb_mode_add('p', pass_enter, pass_leave, pass_keypress, NULL);
 
     /* Prepare the style provider to be used for the clients and completion. */
-    vb.style_provider = gtk_css_provider_get_default();
+    vb.style_provider = gtk_css_provider_new();
 }
 
 /**
@@ -1807,6 +1992,8 @@ static WebKitWebView *webview_new(Client *c, WebKitWebView *webview)
         "signal::ready-to-show", G_CALLBACK(on_webview_ready_to_show), c,
         "signal::web-process-crashed", G_CALLBACK(on_webview_web_process_crashed), c,
         "signal::authenticate", G_CALLBACK(on_webview_authenticate), c,
+        "signal::enter-fullscreen", G_CALLBACK(on_webview_enter_fullscreen), c,
+        "signal::leave-fullscreen", G_CALLBACK(on_webview_leave_fullscreen), c,
         NULL
     );
 
@@ -1834,19 +2021,29 @@ static gboolean on_permission_request(WebKitWebView *webview,
     char *msg = NULL;
 
     if (WEBKIT_IS_GEOLOCATION_PERMISSION_REQUEST(request)) {
-        msg = "Page wants to request your location";
+        char* geolocation_setting = GET_CHAR(c, "geolocation");
+        if (strcmp(geolocation_setting, "ask") == 0) {
+            msg = "access your location";
+        } else if (strcmp(geolocation_setting, "always") == 0) {
+            webkit_permission_request_allow(request);
+            return TRUE;
+        } else if (strcmp(geolocation_setting, "never") == 0) {
+            webkit_permission_request_deny(request);
+            return TRUE;
+        }
     } else if (WEBKIT_IS_USER_MEDIA_PERMISSION_REQUEST(request)) {
         if (webkit_user_media_permission_is_for_audio_device(WEBKIT_USER_MEDIA_PERMISSION_REQUEST(request))) {
-            msg = "Page wants to access the microphone";
+            msg = "access the microphone";
         } else if (webkit_user_media_permission_is_for_video_device(WEBKIT_USER_MEDIA_PERMISSION_REQUEST(request))) {
-            msg = "Page wants to access you webcam";
+            msg = "access you webcam";
         }
     } else {
         return FALSE;
     }
 
     dialog = gtk_message_dialog_new(GTK_WINDOW(c->window), GTK_DIALOG_MODAL,
-            GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, msg);
+            GTK_MESSAGE_QUESTION, GTK_BUTTONS_YES_NO, "Page wants to %s",
+            msg);
 
     gtk_widget_show(dialog);
     result = gtk_dialog_run(GTK_DIALOG(dialog));
@@ -1897,6 +2094,13 @@ static gboolean profileOptionArgFunc(const gchar *option_name,
     return TRUE;
 }
 
+static gboolean autocmdOptionArgFunc(const gchar *option_name,
+        const gchar *value, gpointer data, GError **error)
+{
+    vb.cmdargs = g_slist_append(vb.cmdargs, g_strdup(value));
+    return TRUE;
+}
+
 int main(int argc, char* argv[])
 {
     Client *c;
@@ -1905,10 +2109,13 @@ int main(int argc, char* argv[])
     gboolean ver = FALSE, buginfo = FALSE;
 
     GOptionEntry opts[] = {
-        {"embed", 'e', 0, G_OPTION_ARG_STRING, &winid, "Reparents to window specified by xid", NULL},
+        {"cmd", 'C', 0, G_OPTION_ARG_CALLBACK, (GOptionArgFunc*)autocmdOptionArgFunc, "Ex command run before first page is loaded", NULL},
         {"config", 'c', 0, G_OPTION_ARG_FILENAME, &vb.configfile, "Custom configuration file", NULL},
+        {"embed", 'e', 0, G_OPTION_ARG_STRING, &winid, "Reparents to window specified by xid", NULL},
+        {"incognito", 'i', 0, G_OPTION_ARG_NONE, &vb.incognito, "Run with user data read-only", NULL},
         {"profile", 'p', 0, G_OPTION_ARG_CALLBACK, (GOptionArgFunc*)profileOptionArgFunc, "Profile name", NULL},
         {"version", 'v', 0, G_OPTION_ARG_NONE, &ver, "Print version", NULL},
+        {"no-maximize", 0, 0, G_OPTION_ARG_NONE, &vb.no_maximize, "Do no attempt to maximize window", NULL},
         {"bug-info", 0, 0, G_OPTION_ARG_NONE, &buginfo, "Print used library versions", NULL},
         {NULL}
     };
@@ -1922,12 +2129,12 @@ int main(int argc, char* argv[])
     }
 
     if (ver) {
-        printf("%s, version %s\n\n", PROJECT, VERSION);
+        printf("%s, version %s\n", PROJECT, VERSION);
         return EXIT_SUCCESS;
     }
 
     if (buginfo) {
-        printf("Commit:          %s\n", COMMIT);
+        printf("Version:         %s\n", VERSION);
         printf("WebKit compile:  %d.%d.%d\n",
                 WEBKIT_MAJOR_VERSION,
                 WEBKIT_MINOR_VERSION,
@@ -1974,6 +2181,11 @@ int main(int argc, char* argv[])
 
     c = client_new(NULL);
     client_show(NULL, c);
+
+    /* process the --cmd if this was given */
+    for (GSList *l = vb.cmdargs; l; l = l->next) {
+        ex_run_string(c, l->data, false);
+    }
     if (argc <= 1) {
         vb_load_uri(c, &(Arg){TARGET_CURRENT, NULL});
     } else if (!strcmp(argv[argc - 1], "-")) {
