@@ -80,6 +80,12 @@ var hints = Object.freeze((function(){
         if (removeListener && w) {
             w.removeEventListener("resize", onresize, true);
             w.removeEventListener("scroll", onresize, false);
+            for (i = 0; i < w.frames.length; i++) {
+                try {
+                    w.frames[i].frameElement.contentDocument.removeEventListener("scroll", onresize, false);
+                } catch (ex) {
+                }
+            }
         }
         for (i = 0; i < docs.length; i++) {
             doc = docs[i];
@@ -167,13 +173,10 @@ var hints = Object.freeze((function(){
                 /* create the hint label with number/letters */
                 rect  = e.getClientRects()[0];
                 label = labelTmpl.cloneNode(false);
-                label.setAttribute(
-                    "style", [
-                        "display:none;",
-                        "left:", Math.max((rect.left + offsetX), offsetX), "px;",
-                        "top:", Math.max((rect.top + offsetY), offsetY), "px;"
-                    ].join("")
-                );
+
+                label.style.display = "none";
+                label.style.left    = Math.max(rect.left - 4, 0) + "px";
+                label.style.top     = Math.max(rect.top - 4, 0) + "px";
 
                 /* if hinted element is an image - show title or alt of the image in hint label */
                 /* this allows to see how to filter for the image */
@@ -220,6 +223,10 @@ var hints = Object.freeze((function(){
             /* append the fragment to the document */
             var hDiv = doc.createElement("div");
             hDiv.setAttribute(attr, "container");
+            hDiv.style.position = "fixed";
+            hDiv.style.top      = "0";
+            hDiv.style.left     = "0";
+            hDiv.style.zIndex   = "225000";
             hDiv.appendChild(fragment);
             if (doc.body) {
                 doc.body.appendChild(hDiv);
@@ -341,7 +348,7 @@ var hints = Object.freeze((function(){
                 idx = 0;
             }
         }
-        focusHint(idx);
+        return focusHint(idx);
     }
 
     function fire() {
@@ -416,8 +423,11 @@ var hints = Object.freeze((function(){
         }
         /* get the new active hint */
         if ((activeHint = validHints[newIdx])) {
+            var e = activeHint.e;
             activeHint.focus();
-            mouseEvent(activeHint.e, "mouseover");
+            mouseEvent(e, "mouseover");
+
+            return "OVER:" + (e instanceof HTMLImageElement ? "I:" : "A:") + getSrc(e);
         }
     }
 
@@ -554,6 +564,12 @@ var hints = Object.freeze((function(){
 
             window.addEventListener("resize", onresize, true);
             window.addEventListener("scroll", onresize, false);
+            for (var i = 0; i < window.frames.length; i++) {
+                try {
+                    window.frames[i].frameElement.contentDocument.addEventListener("scroll", onresize, false);
+                } catch (ex) {
+                }
+            }
 
             create();
             return show(true);
