@@ -21,11 +21,13 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <glib.h>
+#include <glib/gstdio.h>
 #include <JavaScriptCore/JavaScript.h>
 #include <pwd.h>
 #include <stdio.h>
 #include <string.h>
 #include <sys/file.h>
+#include <unistd.h>
 
 #include "ascii.h"
 #include "completion.h"
@@ -104,7 +106,7 @@ void util_cleanup(void)
 gboolean util_create_dir_if_not_exists(const char *dirpath)
 {
     if (g_mkdir_with_parents(dirpath, 0755) == -1) {
-        g_critical("Could not create directory '%s': %s", dirpath, strerror(errno));
+        g_critical("Could not create directory '%s': %s", dirpath, g_strerror(errno));
 
         return FALSE;
     }
@@ -131,6 +133,11 @@ gboolean util_create_tmp_file(const char *content, char **file)
         return FALSE;
     }
 
+    if (content == NULL) {
+        close(fp);
+        return TRUE;
+    }
+
     len = strlen(content);
 
     /* write content into temporary file */
@@ -275,7 +282,7 @@ void util_file_prepend_line(const char *file, const char *line,
         }
         g_strfreev(lines);
     }
-    g_file_set_contents(file, new_content->str, -1, NULL);
+    util_file_set_content(file, new_content->str);
     g_string_free(new_content, TRUE);
 }
 
@@ -308,7 +315,7 @@ char *util_file_pop_line(const char *file, int *item_count)
             /* minus one for last empty item and one for popped item */
             count = len - 2;
             new   = g_strjoinv("\n", lines + 1);
-            g_file_set_contents(file, new, -1, NULL);
+            util_file_set_content(file, new);
             g_free(new);
         }
         g_strfreev(lines);
@@ -351,37 +358,76 @@ char *util_get_file_contents(const char *filename, gsize *length)
 }
 
 /**
- * Buil the path from given directory and filename and checks if the file
- * exists. If the file does not exists and the create option is not set, this
- * function returns NULL.
- * If the file exists or the create option was given the full generated path
- * is returned as newly allocated string.
- *
- * The return value must be freed with g_free.
- *
- * @dir:        Directory in which the file is searched.
- * @filename:   Filename to built the absolute path with.
- * @create:     If TRUE, the file is created if it does not already exist.
+ * Atomicly writes contents to given file.
+ * Returns TRUE on success, FALSE otherwise.
  */
-char *util_get_filepath(const char *dir, const char *filename, gboolean create)
+gboolean util_file_set_content(const char *file, const char *contents)
 {
-    char *fullpath;
+    gboolean retval = FALSE;
+    char *tmp_name;
+    int fd, mode;
+    gsize length;
+    struct stat st;
 
-    /* Built the full path out of config dir and given file name. */
-    fullpath = g_build_filename(dir, filename, NULL);
-
-    if (g_file_test(fullpath, G_FILE_TEST_IS_REGULAR)) {
-        return fullpath;
-    } else if (create) {
-        /* If create option was given - create the file. */
-        fclose(fopen(fullpath, "a"));
-        return fullpath;
+    mode = 0600;
+    if (stat(file, &st) == 0) {
+        mode = st.st_mode;
     }
 
-    g_free(fullpath);
-    return NULL;
-}
+    /* Create a temporary file. */
+    tmp_name = g_strconcat(file, ".XXXXXX", NULL);
+    errno    = 0;
+    fd       = g_mkstemp_full(tmp_name, O_RDWR, mode);
+    length   = strlen(contents);
 
+    if (fd == -1) {
+        g_error("Failed to create file %s: %s", tmp_name, g_strerror(errno));
+
+        goto out;
+    }
+
+    /* Write the contents to the temporary file. */
+    while (length > 0) {
+        gssize s;
+        s = write(fd, contents, length);
+        if (s < 0) {
+            if (errno == EINTR) {
+                continue;
+            }
+            g_error("Failed to write to file %s: write() failed: %s",
+                    tmp_name, g_strerror(errno));
+            close(fd);
+            g_unlink(tmp_name);
+
+            goto out;
+        }
+
+        g_assert (s <= length);
+
+        contents += s;
+        length   -= s;
+    }
+
+    if (!g_close(fd, NULL)) {
+        g_unlink(tmp_name);
+        goto out;
+    }
+
+    /* Atomic rename the temporary file into the destination file. */
+    if (g_rename(tmp_name, file) == -1) {
+        g_error("Failed to rename file %s to %s: g_rename() failed: %s",
+                tmp_name, file, g_strerror(errno));
+        g_unlink(tmp_name);
+        goto out;
+    }
+
+    retval = TRUE;
+
+out:
+    g_free(tmp_name);
+
+    return retval;
+}
 
 /**
  * Retrieves the file content as lines.
@@ -397,7 +443,7 @@ char **util_get_lines(const char *filename)
         return NULL;
     }
 
-    if ((content = util_get_file_contents(filename, NULL))) {
+    if (g_file_get_contents(filename, &content, NULL, NULL)) {
         /* split the file content into lines */
         lines = g_strsplit(content, "\n", -1);
         g_free(content);
@@ -410,20 +456,18 @@ char **util_get_lines(const char *filename)
  * based on the lines comparing all chars until the next <tab> char or end of
  * line.
  *
- * @filename:    file to read items from
  * @func:        Function to parse a single line to item.
  * @max_items:   maximum number of items that are returned, use 0 for
  *               unlimited items
  */
-GList *util_file_to_unique_list(const char *filename, Util_Content_Func func,
+GList *util_strv_to_unique_list(char **lines, Util_Content_Func func,
         guint max_items)
 {
-    char *line, **lines;
+    char *line;
     int i, len;
     GList *gl = NULL;
     GHashTable *ht;
 
-    lines = util_get_lines(filename);
     if (!lines) {
         return NULL;
     }
@@ -475,7 +519,6 @@ GList *util_file_to_unique_list(const char *filename, Util_Content_Func func,
         }
     }
 
-    g_strfreev(lines);
     g_hash_table_destroy(ht);
 
     return gl;
@@ -602,6 +645,10 @@ double util_js_result_as_number(WebKitJavascriptResult *result)
  * not expanded char. If no expansion pattern was found, the first char is
  * appended to given GString.
  *
+ * Please note that for a single ~, g_get_home_dir() is used and if a valid
+ * HOME environment variable is set it is preferred than passwd file.
+ * However, for ~user expansion the passwd file is always used.
+ *
  * @input:     String pointer with the content to be parsed.
  * @str:       GString that will be filled with expanded content.
  * @flags      Flags that determine which expansion are processed.
@@ -736,22 +783,36 @@ char *util_sanitize_filename(char *filename)
 /**
  * Strips password from a uri.
  *
- * Return newly allocated string.
+ * Return newly allocated string or NULL.
  */
 char *util_sanitize_uri(const char *uri_str)
 {
     SoupURI *uri;
     char *sanitized_uri;
+    char *for_display;
+
+    if (!uri_str) {
+        return NULL;
+    }
+#if WEBKIT_CHECK_VERSION(2, 24, 0)
+    for_display = webkit_uri_for_display(uri_str);
+    if (!for_display) {
+        for_display = g_strdup(uri_str);
+    }
+#else
+    for_display = g_strdup(uri_str);
+#endif
 
     /* Sanitize the uri only in case there is a @ which might be the indicator
      * for credentials used in uri. */
-    if (!strchr(uri_str, '@')) {
-        return g_strdup(uri_str);
+    if (!strchr(for_display, '@')) {
+        return for_display;
     }
 
-    uri = soup_uri_new(uri_str);
+    uri           = soup_uri_new(for_display);
     sanitized_uri = soup_uri_to_string(uri, FALSE);
     soup_uri_free(uri);
+    g_free(for_display);
 
     return sanitized_uri;
 }
@@ -1049,3 +1110,39 @@ static gboolean match_list(const char *pattern, int patlen, const char *subject)
     }
 }
 
+/**
+ * Get the time span to given string like '1y5dh' (one year and five days and
+ * one hour).
+ */
+GTimeSpan util_string_to_timespan(const char *input)
+{
+    unsigned int multiplier = 0;
+    GTimeSpan sum           = 0;
+    gboolean hasmultiplier  = FALSE;
+
+    while (*input) {
+        if (VB_IS_DIGIT(*input)) {
+            multiplier    = multiplier * 10 + (*input - '0');
+            /* Use separate variable to differentiate between no multiplier
+             * set (is same as 1) or ultiplier of '0'. */
+            hasmultiplier = TRUE;
+        } else {
+            GTimeSpan s = 0;
+            switch (*input) {
+                case 'y': s = 365 * G_TIME_SPAN_DAY; break;
+                case 'w': s = 7 * G_TIME_SPAN_DAY; break;
+                case 'd': s = G_TIME_SPAN_DAY; break;
+                case 'h': s = G_TIME_SPAN_HOUR; break;
+                case 'm': s = G_TIME_SPAN_MINUTE; break;
+                case 's': s = G_TIME_SPAN_SECOND; break;
+            }
+            sum += s * (hasmultiplier ? multiplier : 1);
+            /* Unset multiplier for th epossible next multiplier in input */
+            multiplier    = 0;
+            hasmultiplier = FALSE;
+        }
+        input++;
+    }
+
+    return sum;
+}
